﻿using System;
using OpenBveApi.Runtime;

namespace OpenbveFcmbTrainPlugin
{
    /// <summary>A device controlling the door operation.</summary>
    internal class DoorSelection : Device
    {
        /// <summary>Whether the left doors are closing.</summary>
        private bool LeftDoorsClosing;

        /// <summary>Whether the right doors are closing.</summary>
        private bool RightDoorsClosing;


        /// <summary>Is called when the device state should be updated.</summary>
        /// <param name="train">The current train.</param>
        /// <param name="route">The current route.</param>
        /// <param name="init">Whether the device should initialize.</param>
        /// <param name="elapsedTime">The time elapsed since the previous call.</param>
        internal override void Update(Train train, Route route, bool init, Time elapsedTime)
        {
            if (init)
            {
                // Set the selection state of the doors during initialization
                switch (train.DoorState)
                {
                    case DoorStates.None:
                        RequestedDoorInterlock = DoorInterlockStates.Locked;
                        break;
                    case DoorStates.Both:
                        RequestedDoorInterlock = DoorInterlockStates.Unlocked;
                        break;
                    case DoorStates.Left:
                    case DoorStates.Right:
                        RequestedDoorInterlock = (DoorInterlockStates)train.DoorState;
                        break;
                }
            }

            // Update panel variables for door selection state
            train.Panel[1] = (RequestedDoorInterlock == DoorInterlockStates.Left || RequestedDoorInterlock == DoorInterlockStates.Unlocked) || (((train.DoorState & DoorStates.Left) != 0) && !LeftDoorsClosing) ? 1 : 0;
            train.Panel[2] = (RequestedDoorInterlock == DoorInterlockStates.Right || RequestedDoorInterlock == DoorInterlockStates.Unlocked) || (((train.DoorState & DoorStates.Right) != 0) && !RightDoorsClosing) ? 1 : 0;
        }

        /// <summary>Is called when the state of a key changes.</summary>
        /// <param name="key">The key.</param>
        /// <param name="pressed">Whether the key is pressed or released.</param>
        /// <param name="train">The current train.</param>
        internal override void KeyChange(VirtualKeys key, bool pressed, Train train)
        {
            if (pressed)
            {
                switch (key)
                {
                    case VirtualKeys.G:
                        // Change the selection state of the left doors
                        if (!OpenbveFcmbTrainPlugin.KeysPressed[(int)key])
                        {
                            // Button active only if the doors are not open
                            if ((train.DoorState & DoorStates.Left) == 0 || LeftDoorsClosing)
                            {
                                switch (RequestedDoorInterlock)
                                {
                                    case DoorInterlockStates.Locked:
                                        RequestedDoorInterlock = DoorInterlockStates.Left;
                                        break;
                                    case DoorInterlockStates.Left:
                                        RequestedDoorInterlock = DoorInterlockStates.Locked;
                                        break;
                                    case DoorInterlockStates.Right:
                                        RequestedDoorInterlock = DoorInterlockStates.Unlocked;
                                        break;
                                    case DoorInterlockStates.Unlocked:
                                        RequestedDoorInterlock = DoorInterlockStates.Right;
                                        break;
                                }
                            }
                        }
                        break;
                    case VirtualKeys.H:
                        // Change the selection state of the right doors
                        if (!OpenbveFcmbTrainPlugin.KeysPressed[(int)key])
                        {
                            // Button active only if the doors are not open
                            if ((train.DoorState & DoorStates.Right) == 0 || RightDoorsClosing)
                            {
                                switch (RequestedDoorInterlock)
                                {
                                    case DoorInterlockStates.Locked:
                                        RequestedDoorInterlock = DoorInterlockStates.Right;
                                        break;
                                    case DoorInterlockStates.Left:
                                        RequestedDoorInterlock = DoorInterlockStates.Unlocked;
                                        break;
                                    case DoorInterlockStates.Right:
                                        RequestedDoorInterlock = DoorInterlockStates.Locked;
                                        break;
                                    case DoorInterlockStates.Unlocked:
                                        RequestedDoorInterlock = DoorInterlockStates.Left;
                                        break;
                                }
                            }
                        }
                        break;
                    case VirtualKeys.LeftDoors:
                        if (!OpenbveFcmbTrainPlugin.KeysPressed[(int)key])
                        {
                            // Unselect doors automatically when closing
                            if ((train.DoorState & DoorStates.Left) > 0 && !LeftDoorsClosing && (train.DoorInterlockState == DoorInterlockStates.Left || train.DoorInterlockState == DoorInterlockStates.Unlocked))
                            {
                                LeftDoorsClosing = true;
                                RequestedDoorInterlock = RequestedDoorInterlock == DoorInterlockStates.Left ? DoorInterlockStates.Locked : DoorInterlockStates.Right;
                            }
                            else
                            {
                                LeftDoorsClosing = false;
                            }
                        }
                        break;
                    case VirtualKeys.RightDoors:
                        if (!OpenbveFcmbTrainPlugin.KeysPressed[(int)key])
                        {
                            // Unselect doors automatically when closing
                            if ((train.DoorState & DoorStates.Right) > 0 && !RightDoorsClosing && (train.DoorInterlockState == DoorInterlockStates.Right || train.DoorInterlockState == DoorInterlockStates.Unlocked))
                            {
                                RightDoorsClosing = true;
                                RequestedDoorInterlock = RequestedDoorInterlock == DoorInterlockStates.Right ? DoorInterlockStates.Locked : DoorInterlockStates.Left;
                            }
                            else
                            {
                                RightDoorsClosing = false;
                            }
                        }
                        break;
                }
                // TODO: OpenBVE does not trigger virtual buttons when the AI driver is active, needs changes to main program
            }
        }

        /// <summary>Is called when the state of the doors changes.</summary>
        /// <param name="oldState">The old state of the doors.</param>
        /// <param name="newState">The new state of the doors.</param>
        internal override void DoorChange(DoorStates oldState, DoorStates newState)
        {
            if ((oldState == DoorStates.Both || oldState == DoorStates.Left) && (newState == DoorStates.None || newState == DoorStates.Right))
            {
                LeftDoorsClosing = false;
                // Unselect doors automatically when closed, just in case
                if (RequestedDoorInterlock == DoorInterlockStates.Unlocked || RequestedDoorInterlock == DoorInterlockStates.Left)
                {
                    RequestedDoorInterlock = RequestedDoorInterlock == DoorInterlockStates.Left ? DoorInterlockStates.Locked : DoorInterlockStates.Right;
                }
            }
            if ((oldState == DoorStates.Both || oldState == DoorStates.Right) && (newState == DoorStates.None || newState == DoorStates.Left))
            {
                RightDoorsClosing = false;
                // Unselect doors automatically when closed, just in case
                if (RequestedDoorInterlock == DoorInterlockStates.Unlocked || RequestedDoorInterlock == DoorInterlockStates.Right)
                {
                    RequestedDoorInterlock = RequestedDoorInterlock == DoorInterlockStates.Right ? DoorInterlockStates.Locked : DoorInterlockStates.Left;
                }
            }
        }

        /// <summary>Is called when the device should perform the AI.</summary>
        /// <param name="data">The AI data.</param>
        /// <param name="train">The current train.</param>
        /// <param name="route">The current route.</param>
        internal override void PerformAI(AIData data, Train train, Route route)
        {
            // Select doors to be opened on the current stop, at least 200 meters before the stop point.
            if (route.CurrentStation.PlayerStops() && train.State.Location >= route.CurrentStation.StopPosition - 200 && route.StationState == Route.StationStates.Pending)
            {
                switch (RequestedDoorInterlock)
                {
                    case DoorInterlockStates.Unlocked:
                        if (!route.CurrentStation.OpenLeftDoors)
                        {
                            KeyChange(VirtualKeys.G, true, train);
                            data.Response = AIResponse.Short;
                            KeyChange(VirtualKeys.G, false, train);
                        }
                        if (!route.CurrentStation.OpenRightDoors)
                        {
                            KeyChange(VirtualKeys.H, true, train);
                            data.Response = AIResponse.Short;
                            KeyChange(VirtualKeys.H, false, train);
                        }
                        break;
                    case DoorInterlockStates.Left:
                        if (!route.CurrentStation.OpenLeftDoors)
                        {
                            KeyChange(VirtualKeys.G, true, train);
                            data.Response = AIResponse.Short;
                            KeyChange(VirtualKeys.G, false, train);
                        }
                        if (route.CurrentStation.OpenRightDoors)
                        {
                            KeyChange(VirtualKeys.H, true, train);
                            data.Response = AIResponse.Short;
                            KeyChange(VirtualKeys.H, false, train);
                        }
                        break;
                    case DoorInterlockStates.Right:
                        if (route.CurrentStation.OpenLeftDoors)
                        {
                            KeyChange(VirtualKeys.G, true, train);
                            data.Response = AIResponse.Short;
                            KeyChange(VirtualKeys.G, false, train);
                        }
                        if (!route.CurrentStation.OpenRightDoors)
                        {
                            KeyChange(VirtualKeys.H, true, train);
                            data.Response = AIResponse.Short;
                            KeyChange(VirtualKeys.H, false, train);
                        }
                        break;
                    case DoorInterlockStates.Locked:
                        if (route.CurrentStation.OpenLeftDoors)
                        {
                            KeyChange(VirtualKeys.G, true, train);
                            data.Response = AIResponse.Short;
                            KeyChange(VirtualKeys.G, false, train);
                        }
                        if (route.CurrentStation.OpenRightDoors)
                        {
                            KeyChange(VirtualKeys.H, true, train);
                            data.Response = AIResponse.Short;
                            KeyChange(VirtualKeys.H, false, train);
                        }
                        break;
                }
            }
        }
    }
}

